/*
 * *****************************************************************************************
 *
 * 		Copyright (C) 2014-2021 Gowin Semiconductor Technology Co.,Ltd.
 * 		
 * @file      bring_up.c
 * @author    Embedded Development Team
 * @version   V1.0.0
 * @date      2021-03-24 09:00:00
 * @brief     person_detection flow.
 ******************************************************************************************
 */

/* Includes ---------------------------------------------------------- */
#include <stdbool.h>
#include "GOWIN_M1.h"
#include "ahb_convolution.h"
#include "bring_up.h"
#include "helper.h"
#include "command_responder.h"
#include "tflite_model_parameters.h"
#include "mjb_define.h"

/* Macros ---------------------------------------------------------- */
#define IMG_SIZE 				(96)		//image size
#define USE_PSRAM							//use psram as convolution data buffer

#define TXT_OFFSET				0x0000B000	//HDMI output address

//For Debug
#if defined ENABLE_INPUT_DEBUG || defined ENABLE_OUTPUT_DEBUG
char buff[20];
static char* int2str(int num,char*str,int radix);
#endif
//For Debug

/* Functions ---------------------------------------------------------- */
//Initialization
void setup(void)
{
  SystemInit();				//Initializes system clock
  GPIOInit();				//Initializes GPIO
  I2CInit();				//Initializes I2C
  UartInit();				//Initializes UART0
}

//Convolution loop flow
void loop(void)
{
	AhbTFLiteConvolution_t core;
	
	uint32_t* im_buf = (uint32_t *) AHB2PERIPH_BASE;					//image input
	uint32_t* infer_num = (uint32_t *) (AHB2PERIPH_BASE + TXT_OFFSET);	//hdmi output
	uint32_t im_pix;
	int8_t image[2];

	//Initializes AHB convolution
	AhbTFLiteConvolutionInit(&core, AHB2PERIPH_BASE+0x0000A000);

	while(1)
	{
		AhbTFLiteConvolutionSetMemSelector(&core, 0);
		
		/* ---------------------------------------- Images Provider --------------------------------------*/
		#ifdef FRONTEND_CAMERA_IN
		
		//-------------- Input Debug ------------------//
		#ifdef ENABLE_INPUT_DEBUG
		//input debug information
		UART_SendString(UART0, "\r\n\r\n----------***** Input Debug Information (Camera Image Pixel Data) *****----------\r\n");
		UART_SendString(UART0, "===============================================================================\r\n");
		#endif	//ENABLE_INPUT_DEBUG
		//-------------- Input Debug ------------------//
		
		//input image data
		for (int i=0; i< IMG_SIZE*IMG_SIZE/4; i++)
		{
			im_pix = (*(im_buf+i)) ^ 0x80808080; //Inverted MSB to turn Uint8 to Int8
			
			//---------------- Input Debug ----------------//
			#ifdef ENABLE_INPUT_DEBUG
			//input debug information
			UART_SendChar(UART0, (im_pix&0x000000ff)>>0);	      
			UART_SendChar(UART0, (im_pix&0x0000ff00)>>8);
			UART_SendChar(UART0, (im_pix&0x00ff0000)>>16);
			UART_SendChar(UART0, (im_pix&0xff000000)>>24);
			#endif	//ENABLE_INPUT_DEBUG
			//-------------- Input Debug ------------------//
			
			AhbTFLiteConvolutionLoadImagePixel(&core, (im_pix&0x000000ff)>> 0, 4*i+0);
			AhbTFLiteConvolutionLoadImagePixel(&core, (im_pix&0x0000ff00)>> 8, 4*i+1);	
			AhbTFLiteConvolutionLoadImagePixel(&core, (im_pix&0x00ff0000)>>16, 4*i+2);	
			AhbTFLiteConvolutionLoadImagePixel(&core, (im_pix&0xff000000)>>24, 4*i+3);
		}
		
		//--------------- Input Debug -----------------//
		#ifdef ENABLE_INPUT_DEBUG
		UART_SendString(UART0, "\r\n===============================================================================\r\n");
		#endif	//ENABLE_INPUT_DEBUG
		//--------------- Input Debug -----------------//
		
		#endif	//FRONTEND_CAMERA_IN
		
		/* ------------------------------------- Convolution Inference ----------------------------------*/
		uint32_t wt_bias_mem_offset = 0;
		bool mem_selector = false;
		int wt_mem_offset = 0;
		int bias_mem_offset = 0;
		
		for(int i = 0; i < STAGES; i++)
		{
			uint32_t conv_result_size = CONV_OUT_CH[i] * CONV_OUT_DIM_HEIGHT[i] * CONV_OUT_DIM_WIDTH[i];
     
			core.depthwise = DEPTHWISE[i];
			core.conv_dim_im_in_x = CONV_IN_DIM_WIDTH[i];			//x <--> width
			core.conv_dim_im_in_y = CONV_IN_DIM_HEIGHT[i];		//y <--> height
			core.conv_ch_im_in = CONV_IN_CH[i];
			core.conv_ch_im_out = CONV_OUT_CH[i];
			core.conv_dim_kernel = CONV_KER_DIM_WIDTH[i];			//x = y
			core.conv_padding = CONV_PADDING_WIDTH[i];				//x = y
			core.conv_stride = CONV_STRIDE_WIDTH[i];					//x = y
			core.conv_dilation = DILATION_FACTOR_WIDTH[i];		//x = y
			core.conv_dim_im_out_x = CONV_OUT_DIM_WIDTH[i];
			core.conv_dim_im_out_y = CONV_OUT_DIM_HEIGHT[i];
			core.conv_batch = BATCH[i];
			core.conv_input_offset = IN_OFFSET[i];
			core.conv_filter_offset = WT_OFFSET[i];
			
			core.conv_output_offset = OUT_OFFSET[i];
			core.conv_activation_max = OUT_ACTIVATION_MAX[i];
			core.conv_activation_min = OUT_ACTIVATION_MIN[i];
			core.conv_depth_multiplier = DEPTH_MULTIPLIER[i];
			core.pool_ch_im_in = CONV_OUT_CH[i];
			core.pool_dim_im_in_x = CONV_OUT_DIM_WIDTH[i];
			core.pool_dim_im_in_y = CONV_OUT_DIM_HEIGHT[i];
			core.pool_dim_im_out_x = POOL_OUT_DIM_WIDTH[i];
			core.pool_dim_im_out_y = POOL_OUT_DIM_HEIGHT[i];
			core.pool_dim_kernel =  POOL_KER_DIM_WIDTH[i];	//x = y
			core.pool_padding = POOL_PADDING_WIDTH[i];			//x = y
			core.pool_stride = POOL_STRIDE_WIDTH[i];
			core.pool_en = POOL_EN[i];
			core.pool_mode =  POOL_MODE[i]; 

			int wt_size = CONV_KER_DIM_HEIGHT[i] * CONV_KER_DIM_WIDTH[i] * CONV_OUT_CH[i] * ((core.depthwise)?1:CONV_IN_CH[i]);
			int bias_size = 4*CONV_OUT_CH[i];
			int shift_size = 4*CONV_OUT_CH[i];
			int multiplier_size = 4*CONV_OUT_CH[i];
			core.conv_output_multiplier_offset = wt_size+bias_size;
			core.conv_output_shift_offset = wt_size+bias_size+multiplier_size;
			
			AhbTFLiteConvolutionSetMemSelector(&core, (uint8_t)mem_selector);
			AhbTFLiteConvolutionSetWtBiasMemOffsets(&core, wt_mem_offset, bias_mem_offset);
			AhbTFLiteConvolutionProcess(&core);

			if (i == 27)
			{
				#ifdef USE_PSRAM
				AhbTFLiteConvolutionUnloadImagePixel(&core, (uint16_t)0);
				for(int k = 1; k <= conv_result_size; k++)
				{
					image[k-1] = ((char)AhbTFLiteConvolutionUnloadImagePixel(&core, (uint16_t)k));
				}
				#else
				for(int k = 0; k < conv_result_size; k++)
				{
					image[k] = ((char)AhbTFLiteConvolutionUnloadImagePixel(&core, (uint16_t)k));
				}
				#endif
			}
      
			wt_bias_mem_offset = wt_bias_mem_offset + wt_size + bias_size + shift_size + multiplier_size;
			wt_mem_offset += wt_size;
			bias_mem_offset += bias_size;

			if(!core.pool_en)
			{	
				mem_selector = !mem_selector;
			}
		}
		
		/* ---------------------------------------- Reshape --------------------------------------*/
		//TODO : 
		//Or not to do.
		
		/* ---------------------------------------- Softmax --------------------------------------*/
		//TODO : 
		//Or not to do.

		/* ---------------------------------------- Result --------------------------------------*/
		//image[0] : no person
		//image[1] : person
		//HDMI : person or empty
		//LED2~4(mjb board) and LED1~4(camera board) : as inference indicator
		//LED1(mjb board) : as system indicator
		//UART : person or no person
		command_responder(image, infer_num);
		
		//--------------- Output Debug -----------------//
		#ifdef ENABLE_OUTPUT_DEBUG
		//output debug information
		UART_SendString(UART0, "\r\n\r\n----------***** Output Debug Information (Classification Score) *****----------\r\n");
		UART_SendString(UART0, "===============================================================================\r\n");
		UART_SendString(UART0, "no person score : ");
		UART_SendString(UART0, int2str(image[0],buff,10));    //no person score
		UART_SendString(UART0, "\r\n");
		UART_SendString(UART0, "person score : ");
		UART_SendString(UART0, int2str(image[1],buff,10));    //person score
		UART_SendString(UART0, "\r\n");
		UART_SendString(UART0, "===============================================================================\r\n");
		#endif	//ENABLE_OUTPUT_DEBUG
		//-------------- Output Debug ------------------//
	}
}

//For Debug
#if defined ENABLE_INPUT_DEBUG || defined ENABLE_OUTPUT_DEBUG
//int to string
static char* int2str(int num,char*str,int radix)
{
	char index[]="0123456789ABCDEF";
	unsigned unum;
	int i=0,j,k;
	char temp;

	if(radix==10&&num<0)
	{
		unum=(unsigned)-num;
		str[i++]='-';
	}
	else
	{
		unum=(unsigned)num;
	}

	do
	{
		str[i++]=index[unum%(unsigned)radix];
		unum/=radix;
	}while(unum);
	str[i]='\0';

	if(str[0]=='-')
	{
		k=1;
	}
	else
	{
		k=0;
	}
	
	for(j=k;j<=(i-1)/2;j++)
	{
		temp=str[j];
		str[j]=str[i-1+k-j];
		str[i-1+k-j]=temp;
	}
	
	return str;
}
#endif
//For Debug
